from abc import ABCMeta, abstractmethod
from Algorithm import *
import numpy as np

class Classifier(Algorithm):
    __metaclass__ = ABCMeta
    def __init__(self, dataset):
        # we're gathering the dataset
        self.dataset = dataset
        self.model = []

    @abstractmethod
    def predict():
        pass

    @abstractmethod
    def learn():
        pass



# Adapter for sklearn
from Dataset import *
from Classifier import *
from sklAdapter import *

from sklearn import naive_bayes

class sklGNB(sklAdapter):
    def __init__(self, dataset, configuration = {}):
        Classifier.__init__(self,dataset)
        self.clf = naive_bayes.GaussianNB()


    @classmethod
    def name(cls):
        return 'GNB'



# Adapter for sklearn
from Dataset import *
from Classifier import *
from sklAdapter import *

from sklearn import tree

class sklDTC(sklAdapter):
    def __init__(self, dataset, configuration = {}):
        Classifier.__init__(self,dataset)
        self.clf = tree.DecisionTreeClassifier()


    @classmethod
    def name(cls):
        return 'DTC'



from Sample import *
from utils import getType
import csv
import numpy as np
import random
import re

np.seterr(divide='ignore', invalid='ignore')

SEED = 123
FOLDS = 5
random.seed(SEED)

# DATASET
class Dataset(object):
    def __init__(self, filename=0, resample=0, pureLabels = True):
        #print 'PL %i' % pureLabels
        # Load db
        self.db_name = ""
        self.source_samples = []
        self.train = []
        self.test = []
        self.has_header = False
        self.header = []
        self.classes = {}
        self.resample = resample

        maxPureLabel = 0

        if filename:
            wisepath = re.findall(r"[\w']+", filename)
            self.db_name = wisepath[len(wisepath) - 2]

            with open(filename, 'rb') as file:
                csvDataset = csv.reader(file, delimiter=',')
                for row in csvDataset:
                    if getType(row[0]) == str:
                        self.hasHeader = True
                        self.header = row
                    else:
                        label = row[-1]
                        features = row[0:-1]
                        if pureLabels:
                            self.source_samples.append(Sample(features, int(label)))
                            if int(label) > maxPureLabel:
                                maxPureLabel = int(label)
                        else:
                            if not label in self.classes:
                                self.classes.update({label: len(self.classes)})
                            self.source_samples.append(Sample(features, self.classes[label]))

            if pureLabels:
                for i in xrange(maxPureLabel + 1):
                    self.classes.update({i: i})

            #print self.classes


            self.preAnalyze()

    def setFeatureFilter(self, featureFilter):
        #print 'ORIGINAL %s' % featureFilter
        #nFeatureFilter = [i for i in featureFilter]
        #print 'NEGATED %s' % nFeatureFilter
        self.featureFilter = featureFilter
        for sample in self.source_samples:
            sample.setFeatureFilter(self.featureFilter)
        self.preAnalyze()

    def export(self, filename):
        print 'EXPORTING'
        with open(filename, 'wb') as file:
            csvDataset = csv.writer(file, delimiter=',')
            for sample in self.source_samples:
                row = list(sample.getFeatures())
                row.append(sample.label)
                csvDataset.writerow(row)
                # print sample.

    def injectTrain(self, train):
        self.train = train
        self.samples = self.train
        self.scipyPrepare()

    def fill(self, db_name, source_samples, classes):
        self.db_name = db_name
        self.source_samples = source_samples
        self.classes = classes
        self.preAnalyze(normalisation = False)

    def preAnalyze(self, normalisation = True):
        self.confusion_matrix = []
        # normalize
        if normalisation:
            self.normalize()

        # dumb resampling
        if self.resample:
            random.shuffle(self.source_samples)
            self.source_samples = self.source_samples[0:self.resample]

        # Count features
        self.features = len(self.source_samples[0].getFeatures())

        # Initialize supports
        self.clearSupports()

        # Set base sample set
        self.samples = self.source_samples

        self.prepareCV()
        self.scipyPrepare()

    def scipyPrepare(self):
        # Scipy
        self.X = [x.getFeatures() for x in self.samples]
        self.y = [x.label for x in self.samples]
        self.testX = []

    def prepareCV(self):
        # Prepare CV
        indexes = range(0, len(self.source_samples))
        random.shuffle(indexes)

        self.cv = []
        for index, value in enumerate(indexes):
            self.cv.append((value, index % FOLDS))


    def clearSupports(self):
        for sample in self.source_samples:
            sample.support = np.zeros(len(self.classes))

    def __str__(self):
        return "%s dataset (%i samples, %i features, %i classes)" % (
            self.db_name,
            len(self.source_samples),
            self.features,
            len(self.classes)
        )

    def setCV(self,fold):
        self.samples = []
        self.test = []
        for pair in self.cv:
            if pair[1] == fold:
                self.samples.append(self.source_samples[pair[0]])
            else:
                self.test.append(self.source_samples[pair[0]]);

        if len(self.train):
            self.samples = self.train

        # SKL
        self.scipyPrepare()
        self.testX = [x.getFeatures() for x in self.test]

    def score(self):
        self.confusion_matrix = np.zeros((len(self.classes),len(self.classes))).astype(int)

        for sample in self.test:
            # print "\n# FILLING CONFUSION MATRIX"
            # print sample
            # print sample.label
            # print sample.prediction
            self.confusion_matrix[sample.label,sample.prediction] += 1

        true_positives = np.zeros(len(self.classes)).astype(float)
        false_negatives = np.zeros(len(self.classes)).astype(float)
        false_positives = np.zeros(len(self.classes)).astype(float)
        true_negatives = np.zeros(len(self.classes)).astype(float)

        for pro in xrange(0,len(self.classes)):
            true_positives[pro] += self.confusion_matrix[pro,pro]
            for contra in xrange(0,len(self.classes)):
                if pro == contra: continue
                false_negatives[pro] += self.confusion_matrix[pro,contra]
                false_positives[pro] += self.confusion_matrix[contra,pro]
            true_negatives[pro] = sum(sum(self.confusion_matrix)) - true_positives[pro] - false_negatives[pro] - false_positives[pro]

        sensitivity = true_positives / (true_positives + false_negatives)
        specificity = true_negatives / (false_positives + true_negatives)
        ppv = true_positives / (true_positives + false_positives)
        npv = true_negatives / (true_negatives + false_negatives)
        bac = (sensitivity + specificity) / 2

        acc = sum(true_positives + true_negatives) / sum(true_positives + true_negatives + false_positives + false_negatives)

        scores = \
        {
            'sensitivity': sensitivity.mean(),
            'specificity': specificity.mean(),
            'ppv': ppv.mean(),
            'npv': npv.mean(),
            'accuracy': acc,
            'bac': bac.mean()
        }

        return scores

    def normalize(self):
        example = np.array(self.source_samples[0].getFeatures())
        # Check if there are any NaN's
        for index, value in enumerate(example):
            if np.isnan(value):
                for sample in self.source_samples:
                    if not np.isnan(sample.getFeatures()[index]):
                        example[index] = sample.getFeatures()[index]
                        break

        minimum = np.array(list(example))
        maximum = np.array(list(example))

        for index, sample in enumerate(self.source_samples):
            for index, value in enumerate(sample.getFeatures()):
                if value < minimum[index]:
                    minimum[index] = value
                if value > maximum[index]:
                    maximum[index] = value

        self.normA = minimum
        self.normB = maximum - minimum
        foo = maximum - minimum

        # O tu sobie poradzmy z 0/0
        for index, value in enumerate(foo):
            if value == 0:
                foo[index] = 1

        for sample in self.source_samples:
            for index, feature in enumerate(sample.getFeatures()):
                if not np.isnan(feature):
                    normalizedFeature = (feature - minimum[index]) / foo[index]
                    sample.getFeatures()[index] = normalizedFeature



# Adapter for sklearn
from Dataset import *
from Classifier import *
from sklAdapter import *

from sklearn import neighbors

class sklKNN(sklAdapter):
    def __init__(self, dataset, configuration = { 'k': 5 }):
        Classifier.__init__(self,dataset)
        self.k = configuration['k']
        self.clf = neighbors.KNeighborsClassifier(self.k)

    @classmethod
    def name(cls):
        return 'KNN'



from abc import ABCMeta, abstractmethod
from Algorithm import *
import numpy as np
import random

# To ensure deterministic loop we use randomization with fixed seed.
SEED = 123
random.seed(SEED)

class Ensemble(Algorithm):
    __metaclass__ = ABCMeta
    def __init__(self, dataset):
        # we're gathering the dataset
        self.dataset = dataset

    @abstractmethod
    def predict():
        pass

    @abstractmethod
    def learn():
        pass



# Adapter for sklearn
from Dataset import *
from Classifier import *
from sklAdapter import *

class sklAdapter(Classifier):
    __metaclass__ = ABCMeta
    # === Learning ===
    def learn(self):
        self.clf.fit(self.dataset.X, self.dataset.y)
        pass

    def predict(self):
        predictions = self.clf.predict(self.dataset.testX)
        for idx, val in enumerate(predictions):
            self.dataset.test[idx].prediction = val



from Dataset import *
from Classifier import *

from enum import Enum
import numpy as np
import math
import operator
import png
import functools
import colorsys

class KNN(Classifier):
	def __init__(self, dataset, configuration):
		Classifier.__init__(self,dataset)
		self.k = configuration['k']

	# === Learning ===
	def learn(self):
		pass

	def predict(self):
		for sample in self.dataset.test:
			winners = []
			entry_threshold = 2
			for source in self.dataset.samples:
				distance = np.linalg.norm(sample.features-source.features)
				if distance < entry_threshold:
					winners.append((source, distance))

				if len(winners) > self.k:
					winners.sort(key=lambda distance: distance[1])
					winners = winners[0:self.k]
					entry_threshold = winners[-1][1]

			for winner in winners:
				sample.support[winner[0].label] += 1
			
			sample.decidePrediction()



# IMPORTS
from utils import getType
import csv
import numpy as np
import random

# SAMPLE


class Sample(object):

    def __init__(self, features, label, featureFilter=None):
        # Label is INT
        # Feature vector is float
        width = len(features)
        self.label = label
        self.prediction = 0
        self.support = None
        self.featureFilter = featureFilter

        # Missing values are None
        for index, value in enumerate(features):
            if value == '?':
                features[index] = None

        self.features = np.array(features).astype(np.float)

    def setFeatureFilter(self, featureFilter):
        self.featureFilter = featureFilter

    def decidePrediction(self):
        self.prediction = np.argmax(self.support)

    def getFeatures(self):
        if self.featureFilter is None:
            return self.features
        else:
            return self.features[self.featureFilter]



# Adapter for sklearn
from Dataset import *
from Classifier import *
from sklAdapter import *

from sklearn import neural_network

class sklMLP(sklAdapter):
    def __init__(self, dataset, configuration = {}):
        Classifier.__init__(self,dataset)
        self.clf = neural_network.MLPClassifier(
            solver='lbfgs',
            alpha=1e-5,
            hidden_layer_sizes=(5, 2),
            random_state=1)

    @classmethod
    def name(cls):
        return 'MLP'



from abc import ABCMeta, abstractmethod
import numpy as np

class Algorithm:
    __metaclass__ = ABCMeta
    def quickLoop(self):
        folds = xrange(0,5,1)
        acc = []
        bac = []
        for fold in folds:
            self.dataset.setCV(fold)
            self.learn()
            self.dataset.clearSupports()
            self.predict()
            scores = self.dataset.score()
            acc.append(scores['accuracy'])
            bac.append(scores['bac'])
        macc = np.mean(acc)
        mbac = np.mean(bac)
        return {'acc': macc, 'bac': mbac, 'wacc': acc, 'wbac': bac}

    @classmethod
    def __str__(cls):
        return 'NONE'



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='weles',
    version='0.4.3',
    description='A little bit more abstract Machine Learning library.',
    long_description=readme,
    author='Paweł Ksieniewicz',
    author_email='pawel.ksieniewicz@pwr.edu.pl',
    url='https://github.com/w4k2/weles',
    package_data={'': ['LICENSE']},
    license=license,
    packages=find_packages(exclude=('docs', 'tests', 'README.md'))
)
